library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ElchorSynthesizer is
	port (
		Reset     : in std_logic;
		Clk       : in std_logic;
		MIDIRxIn  : in std_logic;
		BitClkOut : out std_logic;
		LRClkOut  : out std_logic;
		DataOut   : out std_logic
	);
end entity;

architecture A of ElchorSynthesizer is
	component MIDIParser is
		port (
			Reset   : in std_logic;
			Clk     : in std_logic;
			RxIn    : in std_logic;
			TrigOut : out std_logic;
			IncOut  : out std_logic_vector(31 downto 0);
			VelOut  : out std_logic_vector(6 downto 0);
			CC1Out  : out std_logic_vector(6 downto 0);
			CC2Out  : out std_logic_vector(6 downto 0);
			CC3Out  : out std_logic_vector(6 downto 0)
		);
	end component;
	component Oscillator is
		port (
			Clk       : in std_logic;
			Reset     : in std_logic;
			EnableIn  : in std_logic;
			IncIn     : in std_logic_vector(31 downto 0);
			SampleOut : out std_logic_vector(15 downto 0);
			EnableOut : out std_logic
	    );
	end component;
	component EnvelopeGenerator is
		port (
			Clk      : in std_logic;
			Reset    : in std_logic;
			EnableIn : in std_logic;
			KeyIn    : in std_logic;
			ValueOut : out std_logic_vector(31 downto 0)    -- valores entre 0 y 1 en formato de punto fijo Q16.16
	    );
	end component;
	component LJI2SOutput is
		port (
			Clk         : in std_logic;
			Reset       : in std_logic;
			ClkOut      : out std_logic;
			LRClkOut    : out std_logic;
			DataOut     : out std_logic;
			LeftDataIn  : in std_logic_vector(15 downto 0);
			RightDataIn : in std_logic_vector(15 downto 0);
			NextDataOut : out std_logic
	    );
	end component;
	component StateVariableFilter is
		port (
			Reset       : in std_logic;
			Clk         : in std_logic;
			EnableIn    : in std_logic;
			SampleIn    : in std_logic_vector(15 downto 0);
			CutOffIn    : in std_logic_vector(31 downto 0);    -- 0..1  fixed point Q16.16
			ResonanceIn : in std_logic_vector(31 downto 0);    -- 0..1  fixed point Q16.16
			SampleOut   : out std_logic_vector(15 downto 0);
			EnableOut   : out std_logic
		);
	end component;
	signal Trig : std_logic;
	signal OscInc : std_logic_vector(31 downto 0);
	signal CalculateNextSample : std_logic;
	signal O1Out : std_logic_vector(15 downto 0);
	signal SVF1In23 : std_logic_vector(22 downto 0);
	signal SVF1Out : std_logic_vector(15 downto 0);
	signal EG1Out : std_logic_vector(31 downto 0);
	signal SampleData48 : std_logic_vector(47 downto 0);
	signal SampleData16 : std_logic_vector(15 downto 0);
	signal OscillatorEnableOut : std_logic;
	constant SVF_CUTOFF : std_logic_vector(31 downto 0) := x"00002000";
	constant SVF_RESONANCE : std_logic_vector(31 downto 0) := x"00002000";
	signal CC1 : std_logic_vector(6 downto 0);
	signal CC2 : std_logic_vector(6 downto 0);
	signal CC3 : std_logic_vector(6 downto 0);
	signal SvfCutoff : std_logic_vector(31 downto 0);
	signal SvfResonance : std_logic_vector(31 downto 0);
begin
	MP1 : MIDIParser port map (
		Clk => Clk,
		Reset => Reset,
		RxIn => MIDIRxIn,
		TrigOut => Trig,
		IncOut => OscInc,
		CC1Out => CC1,
		CC2Out => CC2,
		CC3Out => CC3
	);
	SvfCutoff    <= x"0000" & CC1 & "000000000";
	SvfResonance <= x"0000" & CC2 & "000000000";
	O1 : Oscillator port map (
		Clk => Clk,
		Reset => Reset,
		EnableIn => CalculateNextSample,
		IncIn => OscInc,
		SampleOut => O1Out,
		EnableOut => OscillatorEnableOut
	);
	EG1 : EnvelopeGenerator port map (
		Clk => Clk,
		Reset => Reset,
		EnableIn => CalculateNextSample,
		KeyIn => Trig,
		ValueOut => EG1Out
	);
	SVF1In23 <= std_logic_vector(signed(O1Out) * signed(CC3));
	SVF1 : StateVariableFilter port map (
		Reset => Reset,
		Clk => Clk,
		EnableIn => OscillatorEnableOut,
		--SampleIn => O1Out,
		SampleIn => SVF1In23(22 downto 7),
		CutOffIn => SvfCutoff,
		ResonanceIn => SvfResonance,
		SampleOut => SVF1Out
	);
	I2S1 : LJI2SOutput port map (
		Clk => Clk,
		Reset => Reset,
		ClkOut => BitClkOut,
		LRClkOut => LRClkOut,
		DataOut => DataOut,
		LeftDataIn => SampleData16,
		RightDataIn => SampleData16,
		NextDataOut => CalculateNextSample
	);
	SampleData48 <= std_logic_vector(signed(SVF1Out) * signed(EG1Out));
	SampleData16 <= SampleData48(31 downto 16);
end architecture;
