library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity LJI2SOutput is
	port (
		Clk         : in std_logic;
		Reset       : in std_logic;
		ClkOut      : out std_logic;
		LRClkOut    : out std_logic;
		DataOut     : out std_logic;
		LeftDataIn  : in std_logic_vector(15 downto 0);
		RightDataIn : in std_logic_vector(15 downto 0);
		NextDataOut : out std_logic
    );
end entity;

architecture A of LJI2SOutput is
	constant CLK_OUT_DIV_BITS : integer := 6;
	signal ClkOutDividerDBus : std_logic_vector((CLK_OUT_DIV_BITS - 1) downto 0);
	signal ClkOutDividerQBus : std_logic_vector((CLK_OUT_DIV_BITS - 1) downto 0);
	constant CLK_OUT_DIV : integer := 23;                 -- (32000000 Hz / 44100 Hz) / 32 bits
	signal ClkOutQBus : std_logic;
	signal ClkOutDBus : std_logic;
	signal BitCounterDBus : std_logic_vector(5 downto 0);
	signal BitCounterQBus : std_logic_vector(5 downto 0);
	signal NextBitCounter : std_logic_vector(5 downto 0);
	signal LRClkOutQBus : std_logic;
	signal LRClkOutDBus : std_logic;
	signal DataDBus : std_logic_vector(31 downto 0);
	signal DataQBus : std_logic_vector(31 downto 0);
begin
	-- clock output divider
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			ClkOutDividerQBus <= ClkOutDividerDBus;
		end if;
	end process;

	ClkOutDividerDBus <= std_logic_vector(to_unsigned(0, CLK_OUT_DIV_BITS)) when ((to_integer(unsigned(ClkOutDividerQBus)) = (CLK_OUT_DIV - 1)) or (Reset = '1')) else
	                     std_logic_vector(to_unsigned(to_integer(unsigned(ClkOutDividerQBus)) + 1, CLK_OUT_DIV_BITS));

	-- bit clock output
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			ClkOutQBus <= ClkOutDBus;
		end if;
	end process;

	ClkOutDBus <= '0' when ((to_integer(unsigned(ClkOutDividerQBus)) = (CLK_OUT_DIV - 1)) or (Reset = '1')) else
	              '1' when  (to_integer(unsigned(ClkOutDividerQBus)) = ((CLK_OUT_DIV / 2) - 1)) else
	              ClkOutQBus;
	ClkOut <= ClkOutQBus;

	-- bit counter
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			BitCounterQBus <= BitCounterDBus;
		end if;
	end process;

	NextBitCounter <= std_logic_vector(to_unsigned(0, 6)) when (to_integer(unsigned(BitCounterQBus)) = 31) else
	                  std_logic_vector(to_unsigned(to_integer(unsigned(BitCounterQBus)) + 1, 6));
	BitCounterDBus <= std_logic_vector(to_unsigned(0, 6)) when (Reset = '1') else
	                  NextBitCounter when (to_integer(unsigned(ClkOutDividerQBus)) = (CLK_OUT_DIV - 1)) else
	                  BitCounterQBus;

	-- left-right clock output
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			LRClkOutQBus <= LRClkOutDBus;
		end if;
	end process;

	LRClkOutDBus <= '1' when ((to_integer(unsigned(BitCounterQBus)) < 16) or (Reset = '1')) else
	                '0';
	LRClkOut <= LRClkOutQBus;

	-- data shift register
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			DataQBus <= DataDBus;
		end if;
	end process;

	DataDBus <= (LeftDataIn & RightDataIn) when ((to_integer(unsigned(ClkOutDividerQBus)) = (CLK_OUT_DIV - 1)) and (to_integer(unsigned(BitCounterQBus)) = 31)) else
	            (DataQBus(30 downto 0) & '0') when ((to_integer(unsigned(ClkOutDividerQBus)) = (CLK_OUT_DIV - 1)) and (to_integer(unsigned(BitCounterQBus)) /= 31)) else
	            DataQBus;
	DataOut <= DataQBus(31);

	NextDataOut <= '1' when ((to_integer(unsigned(ClkOutDividerQBus)) = (CLK_OUT_DIV - 1)) and (to_integer(unsigned(BitCounterQBus)) = 31)) else
	               '0';
end architecture;
