library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity Stage1MIDIParser is
	generic (
		CC1 : integer := 71;    -- filter resonance
		CC2 : integer := 74;    -- filter cutoff
		CC3 : integer := 16     -- filter input gain
	);
	port (
		Reset     : in std_logic;
		Clk       : in std_logic;
		RxIn      : in std_logic;
		KeyOnOut  : out std_logic;
		KeyOffOut : out std_logic;
		KeyOut    : out std_logic_vector(6 downto 0);
		VelOut    : out std_logic_vector(6 downto 0);
		CC1Out    : out std_logic_vector(6 downto 0);
		CC2Out    : out std_logic_vector(6 downto 0);
		CC3Out    : out std_logic_vector(6 downto 0)
	);
end entity;

architecture RTL of Stage1MIDIParser is
	component UartRx is
		generic (
			TIME_COUNTER_BITS   : integer := 11;
			TIME_COUNTER_1BIT   : integer := 1024;   -- 32000000 Hz / 31250 bps = 1024
			TIME_COUNTER_1_5BIT : integer := 1536    -- 1024 * 1.5 = 1536
		);
		port (
			Reset        : in std_logic;
			Clk          : in std_logic;
			RxIn         : in std_logic;
			ByteOut      : out std_logic_vector(7 downto 0);
			ByteOutReady : out std_logic
		);
	end component;
	component NotesRom is
		port (
			AddressIn : in std_logic_vector(6 downto 0);
			DataOut   : out std_logic_vector(31 downto 0)
		);
	end component;
	signal KeyDBus : std_logic_vector(6 downto 0);
	signal KeyQBus : std_logic_vector(6 downto 0);
	signal VelDBus : std_logic_vector(6 downto 0);
	signal VelQBus : std_logic_vector(6 downto 0);
	signal FSMDBus : std_logic_vector(3 downto 0);
	signal FSMQBus : std_logic_vector(3 downto 0);
	signal ByteRx : std_logic_vector(7 downto 0);
	signal ByteRxReady : std_logic;
	signal NoteOnCommandRx : std_logic;
	signal NoteOffCommandRx : std_logic;
	signal DataRx : std_logic;
	signal DataRxIsZero : std_logic;
	signal DataRxIsNotZero : std_logic;
	signal FSMCCDBus : std_logic_vector(2 downto 0);
	signal FSMCCQBus : std_logic_vector(2 downto 0);
	signal CC1DBus : std_logic_vector(6 downto 0);
	signal CC1QBus : std_logic_vector(6 downto 0);
	signal CC2DBus : std_logic_vector(6 downto 0);
	signal CC2QBus : std_logic_vector(6 downto 0);
	signal CC3DBus : std_logic_vector(6 downto 0);
	signal CC3QBus : std_logic_vector(6 downto 0);
	signal ControlChangeCommandRx : std_logic;
	signal DataRxIsCC1 : std_logic;
	signal DataRxIsCC2 : std_logic;
	signal DataRxIsCC3 : std_logic;
begin
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			KeyQBus <= KeyDBus;
		end if;
	end process;

	KeyDBus <= "0000000" when (Reset = '1') else
	           ByteRx(6 downto 0) when ((FSMQBus = "0010") or (FSMQBus = "1000")) else
	           KeyQBus;
	KeyOut <= KeyQBus;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			VelQBus <= VelDBus;
		end if;
	end process;

	VelDBus <= "0000000" when (Reset = '1') else
	           ByteRx(6 downto 0) when (FSMQBus = "0100") else
	           VelQBus;
	VelOut <= VelQBus;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			FSMQBus <= FSMDBus;
		end if;
	end process;

	U1 : UartRx port map (
		Reset => Reset,
		Clk => Clk,
		RxIn => RxIn,
		ByteOut => ByteRx,
		ByteOutReady => ByteRxReady
	);

	NoteOnCommandRx <= '1' when ((ByteRxReady = '1') and (ByteRx(7 downto 4) = "1001")) else
	                   '0';
	NoteOffCommandRx <= '1' when ((ByteRxReady = '1') and (ByteRx(7 downto 4) = "1000")) else
	                    '0';
	DataRx <= '1' when ((ByteRxReady = '1') and (ByteRx(7) = '0')) else
	                   '0';
	DataRxIsZero <= '1' when ((ByteRxReady = '1') and (ByteRx = "00000000")) else
	                '0';
	DataRxIsNotZero <= '1' when ((ByteRxReady = '1') and (ByteRx /= "00000000")) else
	                   '0';

	FSMDBus <= "0000" when ((Reset = '1') or (FSMQBus = "0101") or (FSMQBus = "0110")) else
	           "0001" when ((FSMQBus = "0000") and (NoteOnCommandRx = '1')) else
	           "0010" when ((FSMQBus = "0001") and (DataRx = '1')) else
	           "0011" when (FSMQBus = "0010") else
	           "0100" when ((FSMQBus = "0011") and (DataRxIsNotZero = '1')) else
	           "0101" when (FSMQBus = "0100") else
	           "0110" when (((FSMQBus = "0011") and (DataRxIsZero = '1')) or ((FSMQBus = "1001") and (DataRx = '1'))) else
	           "0111" when ((FSMQBus = "0000") and (NoteOffCommandRx = '1')) else
	           "1000" when ((FSMQBus = "0111") and (DataRx = '1')) else
	           "1001" when (FSMQBus = "1000") else
	           FSMQBus;

	KeyOnOut <= '1' when (FSMQBus = "0101") else
	            '0';
	KeyOffOut <= '1' when (FSMQBus = "0110") else
	             '0';

	-- control change messages

	ControlChangeCommandRx <= '1' when ((ByteRxReady = '1') and (ByteRx(7 downto 4) = "1011")) else
	                          '0';
	DataRxIsCC1 <= '1' when ((ByteRxReady = '1') and (to_integer(unsigned(ByteRx)) = CC1)) else
	               '0';
	DataRxIsCC2 <= '1' when ((ByteRxReady = '1') and (to_integer(unsigned(ByteRx)) = CC2)) else
	               '0';
	DataRxIsCC3 <= '1' when ((ByteRxReady = '1') and (to_integer(unsigned(ByteRx)) = CC3)) else
	               '0';

	FSMCCDBus <= "000" when ((Reset = '1') or (FSMCCQBus = "011") or (FSMCCQBus = "101") or (FSMCCQBus = "111")) else
	             "001" when ((FSMCCQBus = "000") and (ControlChangeCommandRx = '1')) else
	             "010" when ((FSMCCQBus = "001") and (DataRxIsCC1 = '1')) else
	             "011" when ((FSMCCQBus = "010") and (DataRx = '1')) else
	             "100" when ((FSMCCQBus = "001") and (DataRxIsCC2 = '1')) else
	             "101" when ((FSMCCQBus = "100") and (DataRx = '1')) else
	             "110" when ((FSMCCQBus = "001") and (DataRxIsCC3 = '1')) else
	             "111" when ((FSMCCQBus = "110") and (DataRx = '1')) else
	             FSMCCQBus;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			FSMCCQBus <= FSMCCDBus;
		end if;
	end process;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			CC1QBus <= CC1DBus;
		end if;
	end process;

	CC1DBus <= ByteRx(6 downto 0) when (FSMCCQBus = "011") else
	           CC1QBus;
	CC1Out <= CC1QBus;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			CC2QBus <= CC2DBus;
		end if;
	end process;

	CC2DBus <= ByteRx(6 downto 0) when (FSMCCQBus = "101") else
	           CC2QBus;
	CC2Out <= CC2QBus;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			CC3QBus <= CC3DBus;
		end if;
	end process;

	CC3DBus <= ByteRx(6 downto 0) when (FSMCCQBus = "111") else
	           CC3QBus;
	CC3Out <= CC3QBus;
end architecture;
