library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity StateVariableFilter is
	port (
		Reset       : in std_logic;
		Clk         : in std_logic;
		EnableIn    : in std_logic;
		SampleIn    : in std_logic_vector(15 downto 0);
		CutOffIn    : in std_logic_vector(31 downto 0);    -- 0..1  fixed point Q16.16
		ResonanceIn : in std_logic_vector(31 downto 0);    -- 0..1  fixed point Q16.16
		SampleOut   : out std_logic_vector(15 downto 0);
		EnableOut   : out std_logic
	);
end entity;

architecture RTL of StateVariableFilter is
	signal LPDBus : std_logic_vector(31 downto 0);
	signal LPQBus : std_logic_vector(31 downto 0);
	signal HPDBus : std_logic_vector(31 downto 0);
	signal HPQBus : std_logic_vector(31 downto 0);
	signal BPDBus : std_logic_vector(31 downto 0);
	signal BPQBus : std_logic_vector(31 downto 0);
	signal MultOperandA : std_logic_vector(31 downto 0);
	signal MultOperandB : std_logic_vector(31 downto 0);
	signal MultResult64 : std_logic_vector(63 downto 0);
	signal MultResult : std_logic_vector(31 downto 0);
	signal AddOperandB : std_logic_vector(31 downto 0);
	signal AddResult : std_logic_vector(31 downto 0);
	signal NegResonance : std_logic_vector(31 downto 0);
	signal FSMDBus : std_logic_vector(2 downto 0);
	signal FSMQBus : std_logic_vector(2 downto 0);
begin
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			LPQBus <= LPDBus;
		end if;
	end process;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			HPQBus <= HPDBus;
		end if;
	end process;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			BPQBus <= BPDBus;
		end if;
	end process;

	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			FSMQBus <= FSMDBus;
		end if;
	end process;

	NegResonance <= std_logic_vector(to_signed(-to_integer(signed(ResonanceIn)), 32));
	MultOperandA <= CutOffIn when ((FSMQBus = "001") or (FSMQBus = "101")) else
	                NegResonance when (FSMQBus = "011") else
	                std_logic_vector(to_signed(-65536, 32)) when (FSMQBus = "100") else  -- -65536 es -1 en notación Q16.16
	                std_logic_vector(to_signed(0, 32));
	MultOperandB <= LPQBus when (FSMQBus = "100") else
	                BPQBus when ((FSMQBus = "001") or (FSMQBus = "011")) else
	                HPQBus;
	AddOperandB <= LPQBus when (FSMQBus = "001") else
	               BPQBus when (FSMQBus = "101") else
	               HPQBus when ((FSMQBus = "011") or (FSMQBus = "100")) else
	               std_logic_vector(to_signed(to_integer(signed(SampleIn)), 32));
	--MultResult64 <= std_logic_vector(to_signed(to_integer(signed(MultOperandA)) * to_integer(signed(MultOperandB)), 64));
	MultResult64 <= std_logic_vector(signed(MultOperandA) * signed(MultOperandB));
	MultResult <= MultResult64(47 downto 16);
	--AddResult <= std_logic_vector(to_signed(to_integer(signed(MultResult)) + to_integer(signed(AddOperandB)), 32));
	AddResult <= std_logic_vector(signed(MultResult) + signed(AddOperandB));
	LPDBus <= std_logic_vector(to_signed(0, 32)) when (Reset = '1') else
	          AddResult when (FSMQBus = "001") else
	          LPQBus;
	HPDBus <= std_logic_vector(to_signed(0, 32)) when (Reset = '1') else
	          AddResult when ((FSMQBus = "011") or (FSMQBus = "100") or (FSMQBus = "010")) else
	          HPQBus;
	BPDBus <= std_logic_vector(to_signed(0, 32)) when (Reset = '1') else
	          AddResult when (FSMQBus = "101") else
	          BPQBus;

	-- fsm
	--    LP += cutoff * BP
	--    HP = in - (resonance * BP) - LP
	--    BP += cutoff * HP
	FSMDBus <= "000" when ((Reset = '1') or (FSMQBus = "110")) else       --       MultOperandA   MultOperandB   AddOperandB
	           "001" when ((FSMQBus = "000") and (EnableIn = '1')) else   -- LP := cutoff       * BP           + LP
	           "010" when (FSMQBus = "001") else                          -- HP := 0            * x            + IN
	           "011" when (FSMQBus = "010") else                          -- HP := -reso        * BP           + HP
	           "100" when (FSMQBus = "011") else                          -- HP := -1           * LP           + HP
	           "101" when (FSMQBus = "100") else                          -- BP := cutoff       * HP           + BP
	           "110" when (FSMQBus = "101") else
	           "000";
	EnableOut <= '1' when (FSMQBus = "110") else
	             '0';
	SampleOut <= std_logic_vector(to_signed(-32768, 16)) when (to_integer(signed(LPQBus)) < -32768) else
	             std_logic_vector(to_signed(32767, 16)) when (to_integer(signed(LPQBus)) > 32767) else
	             LPQBus(15 downto 0);
end architecture;
