library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity UartRx is
	generic (
		TIME_COUNTER_BITS   : integer := 11;
		TIME_COUNTER_1BIT   : integer := 1024;
		TIME_COUNTER_1_5BIT : integer := 1536
	);
	port (
		Reset        : in std_logic;
		Clk          : in std_logic;
		RxIn         : in std_logic;
		ByteOut      : out std_logic_vector(7 downto 0);
		ByteOutReady : out std_logic
	);
end entity;

architecture RTL of UartRx is
	signal ShiftRegisterDBus : std_logic_vector(8 downto 0);
	signal ShiftRegisterQBus : std_logic_vector(8 downto 0);
	signal ShiftRegisterMuxSel : std_logic;
	signal TimeCounterDBus : std_logic_vector((TIME_COUNTER_BITS - 1) downto 0);
	signal TimeCounterQBus : std_logic_vector((TIME_COUNTER_BITS - 1) downto 0);
	signal TimeCounterMuxSel : std_logic_vector(1 downto 0);
	signal TimeCounterExpired : std_logic;
	signal BitCounterDBus : std_logic_vector(3 downto 0);
	signal BitCounterQBus : std_logic_vector(3 downto 0);
	signal BitCounterMuxSel : std_logic_vector(1 downto 0);
	signal BitCounterExpired : std_logic;
	signal FSMDBus : std_logic_vector(2 downto 0);
	signal FSMQBus : std_logic_vector(2 downto 0);
begin
	-- time counter
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			TimeCounterQBus <= TimeCounterDBus;
		end if;
	end process;

	TimeCounterDBus <= std_logic_vector(to_unsigned(TIME_COUNTER_1BIT, TIME_COUNTER_BITS)) when (TimeCounterMuxSel = "11") else
	                   std_logic_vector(to_unsigned(TIME_COUNTER_1_5BIT, TIME_COUNTER_BITS)) when (TimeCounterMuxSel = "10") else
	                   std_logic_vector(to_unsigned(to_integer(unsigned(TimecounterQBus)) - 1, TIME_COUNTER_BITS)) when (TimeCounterMuxSel = "01") else
	                   TimeCounterQBus;
	TimeCounterExpired <= '1' when (to_integer(unsigned(TimeCounterQBus)) = 0) else
	                      '0';

	-- bit counter
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			BitCounterQBus <= BitCounterDBus;
		end if;
	end process;

	BitCounterDBus <= std_logic_vector(to_unsigned(0, 4)) when (BitCounterMuxSel = "10") else
	                  std_logic_vector(to_unsigned(to_integer(unsigned(BitCounterQBus)) + 1, 4)) when (BitCounterMuxSel = "01") else
	                  BitCounterQBus;
	BitCounterExpired <= '1' when (to_integer(unsigned(BitCounterQBus)) = 8) else
	                     '0';

	-- shift register
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			ShiftRegisterQBus <= ShiftRegisterDBus;
		end if;
	end process;

	ShiftRegisterDBus <= (RxIn & ShiftRegisterQBus(8 downto 1)) when (ShiftRegisterMuxSel = '1') else
	                     ShiftRegisterQBus;
	ByteOut <= ShiftRegisterQBus(7 downto 0);

	-- state machine
	process (Clk)
	begin
		if (Clk'event and (Clk = '1')) then
			FSMQBus <= FSMDBus;
		end if;
	end process;

	-- state machine (next state logic)
	FSMDBus <= "000" when (Reset = '1') else
	           "001" when ((FSMQBus = "000") and (RxIn = '0')) else
	           "010" when ((FSMQBus = "001") or ((FSMQBus = "010") and (TimeCounterExpired = '0'))) else
	           "011" when (((FSMQBus = "010") and (TimeCounterExpired = '1')) or ((FSMQBus = "100") and (TimeCounterExpired = '1'))) else
	           "100" when (((FSMQBus = "011") and (BitCounterExpired = '0')) or ((FSMQBus = "100") and (TimeCounterExpired = '0'))) else
	           "101" when ((FSMQBus = "011") and (BitCounterExpired = '1')) else
	           "000";

	-- state machine (output logic)
	TimeCounterMuxSel <= "11" when (FSMQBus = "011") else
	                     "10" when (FSMQBus = "001") else
	                     "01" when ((FSMQBus = "010") or (FSMQBus = "100")) else
	                     "00";
	BitCounterMuxSel <= "10" when (FSMQBus = "001") else
	                    "01" when (FSMQBus = "011") else
	                    "00";
	ShiftRegisterMuxSel <= '1' when (FSMQBus = "011") else
	                       '0';
	ByteOutReady <= '1' when (FSMQBus = "101") else
	                '0';
end architecture;
